/* FTP extention for TCP NAT alteration. */
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_expect.h"
#include "NAT/ip_nat_helper.h"

EXPORT_NO_SYMBOLS;

static size_t ftp_offset;

#if 0
#define DEBUGFTP printk
#else
#define DEBUGFTP(format, args...)
#endif

/* 75 seconds before we give up. */
#define FTP_TIMEOUT (75*HZ)

/* We keep track of whether it's NAT or RNAT, where the last SYN
   correction was, and the SYN offsets before and after that
   correction. */

enum ftp_mangle_type
{
	FTP_UNINITIALIZED,
	FTP_NONE,
	FTP_NAT,
	FTP_RNAT
};

struct ftp_info
{
	enum ftp_mangle_type mangle_type;
	u_int32_t new_ip;
	u_int32_t syn_correction_pos;
	int32_t syn_offset_before, syn_offset_after;
};

/* This is interesting.  We simply use the port given us by the client
   or server.  In practice it's extremely unlikely to clash; if it
   does, the rule won't be able to get a unique tuple and will drop
   the packets. */
static int
mangle_packet(struct sk_buff **pskb, 
	      const char *preamble,
	      u_int32_t newip, 
	      struct ip_conntrack_expect *expected,
	      const char *postamble,
	      struct ftp_info *ftp)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	char *data = (void *)tcph + tcph->doff*4;
	size_t newlen, newtcplen;
	char buffer[25]; /* nnn,nnn,nnn,nnn,nnn,nnn */
	struct ip_conntrack_tuple newtuple;
	/* This will rewrite it back to original */
	struct ip_conntrack_manip manip = { expected->tuple.dst.ip, 
					    { expected->tuple.dst.u.tcp.port },
					    0 };

	/* New tuple: destination IP will be different. */
	newtuple = expected->tuple;
	newtuple.dst.ip = newip;

	sprintf(buffer, "%u,%u,%u,%u,%u,%u", 
		IP_PARTS(newip), ntohs(expected->tuple.dst.u.tcp.port)>>8,
		ntohs(expected->tuple.dst.u.tcp.port) & 0xFF);

	DEBUGFTP("Mapping tcph %p to new address: `%s'\n", tcph, buffer);
	newtcplen = strlen(preamble) + strlen(buffer) + strlen(postamble)
		+ tcph->doff*4;
	newlen = iph->ihl*4 + newtcplen;

	if (newlen > (*pskb)->len + skb_tailroom(*pskb)) {
		printk("Call rusty: need skb_realloc_tailroom!\n");
		return 0;
	}

	/* SYN adjust.  If it's uninitialized, or this is after last
           correction, record it: we don't handle more than one
           adjustment in the window, but do deal with common case of a
           retransmit. */
	if (ftp->syn_offset_before == ftp->syn_offset_after
	    || before(ftp->syn_correction_pos, ntohl(tcph->seq))) {
		ftp->syn_correction_pos = ntohl(tcph->seq);
		ftp->syn_offset_before = ftp->syn_offset_after;
		ftp->syn_offset_after = (int32_t)
			ftp->syn_offset_before + newlen - (*pskb)->len;
	}

	/* Rewrite packet. */
	if (newlen > (*pskb)->len) {
		DEBUGFTP("ip_nat_ftp: Extending packet by %u to %u bytes\n",
			 newlen - (*pskb)->len, newlen);
		skb_put(*pskb, newlen - (*pskb)->len);
	} else {
		DEBUGFTP("ip_nat_ftp: Shrinking packet from %u to %u bytes\n",
			 (*pskb)->len, newlen);
		skb_trim(*pskb, newlen);
	}

	memcpy(data, preamble, strlen(preamble));
	memcpy(data + strlen(preamble), buffer, strlen(buffer));
	memcpy(data + strlen(preamble) + strlen(buffer), postamble, 
	       strlen(postamble));
	iph->tot_len = htons(newlen);
	(*pskb)->csum = csum_partial((char *)tcph + tcph->doff*4,
				     newtcplen - tcph->doff*4, 0);
	tcph->check = 0;
	tcph->check = tcp_v4_check(tcph, newtcplen, iph->saddr, iph->daddr,
				   csum_partial((char *)tcph, tcph->doff*4,
						(*pskb)->csum));
	ip_send_check(iph);

	/* Tell NAT we need a fixed mapping for this expected conn. */
	if (!ip_nat_expect(&newtuple, &manip, IP_NAT_MANIP_DST, FTP_TIMEOUT)) {
		DEBUGP("ftp: ip_nat_expect failed!\n");
		return 0;
	}

	/* Now change conntrack's expectations. */
	expected->tuple.dst.ip = newip;
	return 1;
}

/* Grrr... SACK.  Fuck me even harder.  Don't want to fix it on the
   fly, so blow it away. */
static void 
delete_sack(struct sk_buff *skb, struct tcphdr *tcph)
{
	unsigned int i;
	u_int8_t *opt = (u_int8_t *)tcph;

	DEBUGFTP("Seeking SACKPERM in SYN packet (doff = %u).\n",
		 tcph->doff * 4);
	for (i = sizeof(struct tcphdr); i < tcph->doff * 4;) {
		DEBUGFTP("%u ", opt[i]);
		switch (opt[i]) {
		case TCPOPT_NOP:
		case TCPOPT_EOL:
			i++;
			break;

		case TCPOPT_SACK_PERM:
			goto found_opt;

		default:
			/* Worst that can happen: it will take us over. */
			i += opt[i+1] ?: 1;
		}
	}
	DEBUGFTP("\n");
	return;

 found_opt:
	DEBUGFTP("\n");
	DEBUGFTP("Found SACKPERM at offset %u.\n", i);

	/* Must be within TCP header, and valid SACK perm. */
	if (i + opt[i+1] <= tcph->doff*4 && opt[i+1] == 2) {
		/* Replace with NOPs. */
		tcph->check
			= ip_nat_cheat_check(*((u_int16_t *)(opt + i))^0xFFFF,
					     0, tcph->check);
		opt[i] = opt[i+1] = 0;
	}
	else DEBUGFTP("Something wrong with SACK_PERM.\n");
}

static unsigned int help(struct ip_conntrack_tuple_hash *h,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info conntrackinfo,
			 struct sk_buff **pskb)
{
#if 0
	printk("ip_nat_ftp helper called.\n");
#else
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	struct ftp_info *ftp = (void *)h->ctrack + ftp_offset;
	u_int32_t newack;
	u_int32_t *adjustme;
	int32_t offset;

	if (ftp->mangle_type == FTP_UNINITIALIZED) {
		IP_CONNTRACK_ASSERT(DIRECTION(h) == 0);
		if (info->dirinfo[0].num_manips == 0) {
			DEBUGFTP("FTP_UNINITIALIZED: num_manips == 0\n");
			ftp->mangle_type = FTP_NONE;
		/* NAT occurs on initial packets on POST_ROUTING hook */
		} else if (info->dirinfo[0].hook == NF_IP_POST_ROUTING) {
			DEBUGFTP("FTP_UNINITIALIZED: NAT\n");
			ftp->mangle_type = FTP_NAT;
			/* New IP: same as where other replies go. */
			ftp->new_ip = h->ctrack->tuplehash[1].tuple.dst.ip;
		} else {
			DEBUGFTP("FTP_UNINITIALIZED: RNAT\n");
			ftp->mangle_type = FTP_RNAT;
			/* New IP: same as where other initial packets go. */
			ftp->new_ip = h->ctrack->tuplehash[0].tuple.dst.ip;
		}
	}

	/* Delete SACK_OK on initial TCP SYNs. */
	if (tcph->syn && !tcph->ack)
		delete_sack(*pskb, tcph);

	DEBUGFTP("ip_nat_ftp: Conntrackinfo = %u, mangle_type = %s\n",
		 conntrackinfo, ftp->mangle_type == FTP_NAT ? "FTP_NAT"
		 : (ftp->mangle_type == FTP_RNAT ? "FTP_RNAT"
		    : (ftp->mangle_type == FTP_NONE ? "FTP_NONE"
		       : "UNKNOWN")));


	/* FIXME MAYBE: It's possible that some servers or clients
	   will get upset if the connection to them comes from a
	   different IP address.  We should do an expected mapping for
	   those, too, in that case. --RR */

	/* FIXME: Yep.  Most servers do anti-spoofing: they won't send
	   a data connection out an interface other than the one which
	   the control connection is on. --RR */
	if (conntrackinfo == IP_CT_ESTABLISHED_SPECIFIES_RELATED) {
		if (DIRECTION(h) == 0 && ftp->mangle_type == FTP_NAT) {
			/* Client gives PORT command */
			DEBUGFTP("ip_nat_ftp: Client PORT command!\n");
			if (!mangle_packet(pskb, "PORT ", ftp->new_ip, 
					   &h->ctrack->expected, "\r\n", ftp)) 
				return NF_DROP;
		} else if (DIRECTION(h) == 1 && ftp->mangle_type == FTP_RNAT) {
			/* Server gives PASV response */
			DEBUGFTP("ip_nat_ftp: Server PASV response!\n");
			if (!mangle_packet(pskb,
					   "227 Entering Passive Mode (", 
					   ftp->new_ip, &h->ctrack->expected,
					   ")\r\n", ftp))
				return NF_DROP;
		}
	}

	if ((DIRECTION(h) == 1 && ftp->mangle_type == FTP_NAT)
	    || (DIRECTION(h) == 0 && ftp->mangle_type == FTP_RNAT)) {
		/* Adjust ACK numbers coming from server (NAT), or
		   client (RNAT). */
		adjustme = &tcph->ack_seq;
		if (after(ntohl(*adjustme), ftp->syn_correction_pos))
			offset = -ftp->syn_offset_after;
		else
			offset = -ftp->syn_offset_before;
		DEBUGFTP("Adjusting ACK number ");
	} else {
		/* Adjust sequence numbers coming from client (NAT), or
		   server (RNAT). */
		adjustme = &tcph->seq;
		if (after(ntohl(*adjustme), ftp->syn_correction_pos))
			offset = ftp->syn_offset_after;
		else
			offset = ftp->syn_offset_before;
		DEBUGFTP("Adjusting SEQ number ");
	}

	newack = htonl(ntohl(*adjustme) + offset);

	DEBUGFTP("from %lu to %lu\n", ntohl(*adjustme), ntohl(newack));
	tcph->check = ip_nat_cheat_check(~*adjustme, newack, tcph->check);
	*adjustme = newack;
#endif
	return NF_ACCEPT;
}

struct ip_nat_helper ftp = { { NULL, NULL }, IPPROTO_TCP, __constant_htons(21),
			     help };

extern struct module *ip_conntrack_ftp;

int init_module(void)
{
	printk("ip_nat_ftp v%s\n", NETFILTER_VERSION);
	ftp_offset = struct_reserve(ip_conntrack, sizeof(struct ftp_info));
	__MOD_INC_USE_COUNT(ip_conntrack_ftp);
	return ip_nat_helper_register(&ftp);
}

void cleanup_module(void)
{
	struct_unreserve(ip_conntrack, ftp_offset, sizeof(struct ftp_info));
	__MOD_DEC_USE_COUNT(ip_conntrack_ftp);
	ip_nat_helper_unregister(&ftp);
}
