/* This file contains all the functions required for the standalone
   ip_nat module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/spinlock.h>

#include "conntrack/struct_reserve.h"
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_rule.h"
#include "NAT/ip_nat_protocol.h"
#include "NAT/ip_nat_core.h"
#include "NAT/ip_nat_helper.h"
#include "NAT/ip_nat_expect.h"
#include "common/listhelp.h"

static unsigned int 
alloc_null_binding(struct ip_conntrack *conntrack,
		   const struct net_device *device)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED). */
	struct ip_nat_multi_range mr = { 1, 
					 { { IP_NAT_RANGE_MAP_IPS, 
					     conntrack->tuplehash[0].tuple.src.ip,
					     conntrack->tuplehash[0].tuple.src.ip,
					     { 0 }, { 0 } } } };

	return ip_nat_setup_info(conntrack, IP_NAT_MANIP_SRC, &mr, device);
}

static unsigned int 
init_conntrack(struct ip_conntrack *conntrack,
	       const struct net_device *device)
{
	struct ip_nat_rule *rule;
	struct ip_nat_info *info = (void *)conntrack + ip_nat_offset;
	unsigned int ret;

	rule = find_rule(&conntrack->tuplehash[0].tuple);

	/* If there's a rule which matches, set up the translation. */
	if (rule) {
		if (rule->mtype)
			ret = rule->mtype->setup(conntrack, rule->maniptype,
						 &rule->mr, device, 
						 rule->moptions);
		else ret = ip_nat_setup_info(conntrack, rule->maniptype,
					     &rule->mr, device);
	} else {
		/* Set up null or pseudo-null binding. */
		ret = alloc_null_binding(conntrack, device);
	}

	if (ret == NF_ACCEPT) {
		place_in_hashes(conntrack);
		info->initialized = 1;
	}

	return ret;
}

static unsigned int 
ip_nat_fn(unsigned int hooknum,
	  struct sk_buff **pskb,
	  const struct net_device *in,
	  const struct net_device *out)
{
	struct ip_conntrack_tuple_hash *h;
	enum ip_conntrack_info ctinfo;
	struct ip_nat_info *info;
	unsigned long flags;
	unsigned int ret = NF_ACCEPT;

	/* FIXME: One day, fill in properly. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN | NFC_ALTERED;

	/* If we had a hardware checksum before, it's now invalid */
	(*pskb)->ip_summed = CHECKSUM_NONE;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	ctinfo = ip_conntrack_findinfo(*pskb, &h);

	switch (ctinfo) {
	case IP_CT_INVALID:
		DEBUGP("skb %p = IP_CT_INVALID\n", *pskb);
		ret = NF_ACCEPT;
		goto out;
		
	case IP_CT_RELATED:
		if ((*pskb)->nh.iph->protocol == IPPROTO_ICMP) {
			icmp_reply_translation(*pskb, h->ctrack, DIRECTION(h));
			ret = NF_ACCEPT;
			goto out;
		}
		/* Fall thru... */
	case IP_CT_NEW:
		info = (void *)h->ctrack + ip_nat_offset;

		/* Seen it before?  This can happen for loopback, retrans,
		   or local packets.. */
		if (info->initialized) {
			break;
		}

		/* Device == NULL means a local packet, since it the hook
		   is either NF_IP_PRE_ROUTING or NF_IP_LOCAL_IN. */
		ret = init_conntrack(h->ctrack, in);
		if (ret != NF_ACCEPT) 
			goto out;
		break;

	default:
		/* ESTABLISHED / ESTABLISHED_SPECIFIES_RELATED */ 
		info = (void *)h->ctrack + ip_nat_offset;
	}

	if (info->dirinfo[DIRECTION(h)].hook == hooknum)
		ret = do_binding(h, ctinfo, pskb);

	/* Special case: catch source-manip'ed packet coming in */
	if (hooknum == NF_IP_LOCAL_IN 
	    && info->dirinfo[0].hook == NF_IP_POST_ROUTING)
		ip_nat_undo(h->ctrack);

 out:
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	return ret;
}

/* Packet filtering is priority 0: for local_in, we want to be after
   it, for local_out, we want to be before.  And after connection tracking.
*/
static struct nf_hook_ops ip_nat_in_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_PRE_ROUTING, 0 };
static struct nf_hook_ops ip_nat_out_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_POST_ROUTING, 0 };
static struct nf_hook_ops ip_nat_local_in_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_LOCAL_IN, 100 };
static struct nf_hook_ops ip_nat_local_out_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_LOCAL_OUT, -100 };

/* Protocol registration. */
int ip_nat_protocol_register(struct ip_nat_protocol *proto)
{
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	list_prepend(&protos, proto);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	MOD_INC_USE_COUNT;
	return 0;
}

void ip_nat_protocol_unregister(struct ip_nat_protocol *proto)
{
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	LIST_DELETE(&protos, proto);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	MOD_DEC_USE_COUNT;
}

static int init_or_cleanup(int init)
{
	int ret = 0;
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	if (!init) goto cleanup;

	ret = ip_nat_rule_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_nothing;
	}
	ret = ip_nat_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_rule_init;
	}

	ret = nf_register_hook(&ip_nat_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nat;
	}
	ret = nf_register_hook(&ip_nat_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	ret = nf_register_hook(&ip_nat_local_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local in hook.\n");
		goto cleanup_outops;
	}
	ret = nf_register_hook(&ip_nat_local_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local out hook.\n");
		goto cleanup_localinops;
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	__MOD_INC_USE_COUNT(ip_conntrack_module);
	return ret;

 cleanup:
	__MOD_DEC_USE_COUNT(ip_conntrack_module);
	nf_unregister_hook(&ip_nat_local_out_ops);
 cleanup_localinops:
	nf_unregister_hook(&ip_nat_local_in_ops);
 cleanup_outops:
	nf_unregister_hook(&ip_nat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ip_nat_in_ops);
 cleanup_nat:
	ip_nat_cleanup();
 cleanup_rule_init:
	ip_nat_rule_cleanup();
 cleanup_nothing:
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	return ret;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_nat_setup_info);
EXPORT_SYMBOL(ip_nat_register_mapping_type);
EXPORT_SYMBOL(ip_nat_unregister_mapping_type);
EXPORT_SYMBOL(ip_nat_helper_register);
EXPORT_SYMBOL(ip_nat_helper_unregister);
EXPORT_SYMBOL(ip_nat_expect);
EXPORT_SYMBOL(ip_nat_cheat_check);

