/* Masquerade.  Simple mapping which alters range to a local IP address
   (depending on route). */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>
#include "NAT/ip_nat_rule.h"

EXPORT_NO_SYMBOLS;

/* Offset in conntrack for masquerading. */
static size_t masq_offset;

static int
masquerade_check(const struct ip_nat_rule *rule)
{
	if (rule->maniptype != IP_NAT_MANIP_SRC) {
		DEBUGP("ipnat_rule_masquerade_check: must be SRC manip.\n");
		return 0;
	}

	return 1;
}

static unsigned int 
masquerade_setup(struct ip_conntrack *conntrack,
		 enum ip_nat_manip_type maniptype,
		 struct ip_nat_multi_range *mr,
		 const struct net_device *device,
		 void *moptions)
{
	unsigned int ret;
	unsigned int i;
	u_int32_t newsrc;
	struct rtable *rt;
	struct ip_nat_multi_range *newrange;
	int *masq_index = (void *)conntrack + masq_offset;

	/* Pass 0 instead of saddr, since it's going to be changed
	   anyway. */
	/* This should really be ip_route_input, but we don't have skb. */
	if (ip_route_output(&rt, conntrack->tuplehash[0].tuple.dst.ip, 0, 0, 0)
	    != 0) {
		DEBUGP("ipnat_rule_masquerade: Can't reroute.\n");
		return NF_DROP;
	}

	newsrc = inet_select_addr(rt->u.dst.dev, rt->rt_gateway,
				  RT_SCOPE_UNIVERSE);
	*masq_index = rt->u.dst.dev->ifindex;
	ip_rt_put(rt);

	newrange = kmalloc(sizeof(struct ip_nat_multi_range)
			   + sizeof(struct ip_nat_range) * (mr->rangesize-1),
			   GFP_ATOMIC);
	if (!newrange) {
		DEBUGP("Masquerade: oom.\n");
		return NF_DROP;
	}

	/* Transfer from original range. */
	memcpy(newrange, mr,
	       sizeof(struct ip_nat_multi_range)
	       + sizeof(struct ip_nat_range) * (mr->rangesize-1));

	/* Substitute IPs for new IP dst. */
	for (i = 0; i < mr->rangesize; i++) {
		newrange->range[i].flags |= IP_NAT_RANGE_MAP_IPS;
		newrange->range[i].max_ip = newrange->range[i].min_ip
			= newsrc;
	}

	/* Hand modified range to generic setup. */
	ret = ip_nat_setup_info(conntrack, maniptype, newrange, device);
	kfree(newrange);
	return ret;
}

int masq_device_event(struct notifier_block *this,
		      unsigned long event,
		      void *ptr)
{
	struct net_device *dev = ptr;
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	if (event == NETDEV_DOWN) {
		/* Device was downed.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		unsigned int i;

		IP_CONNTRACK_ASSERT(dev->ifindex != 0);
		for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
			struct list_head *l;

			for (l = ip_conntrack_hash[i].next;
			     l != &ip_conntrack_hash[i];
			     ) {
				struct list_head *next = l->next;
				int *masq_index = (void *)l + masq_offset;

				/* Non-masqs will be zero-initialized. */
				if (*masq_index == dev->ifindex)
					ip_conntrack_delete((struct 
							     ip_conntrack *)l);
				l = next;
			}
		}
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	return NOTIFY_DONE;
}

static struct notifier_block masq_dev_notifier = {
	masq_device_event,
	NULL,
	0
};


static struct ip_nat_mapping_type mapping_masquerade
= { { NULL, NULL }, "masquerade", &__this_module,
    masquerade_check,
    masquerade_setup,
    0
};

int init_module(void)
{
	int ret;

	printk("ip_nat_map_masquerade v%s\n", NETFILTER_VERSION);
	ret = ip_nat_register_mapping_type(&mapping_masquerade);

	if (ret == 0) {
		masq_offset = struct_reserve(ip_conntrack, sizeof(int));
		/* Register for device down reports */
		register_netdevice_notifier(&masq_dev_notifier);
	}

	return ret;
}

void cleanup_module(void)
{
	ip_nat_unregister_mapping_type(&mapping_masquerade);
	struct_unreserve(ip_conntrack, masq_offset, sizeof(int));
	unregister_netdevice_notifier(&masq_dev_notifier);
}
