#ifndef _IP_CONNTRACK_H
#define _IP_CONNTRACK_H
/* Connection state tracking for netfilter.  This is separated from,
   but required by, the NAT layer; it can also be used by an iptables
   extention. */

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

/* A `tuple' is a structure containing the information to uniquely
  identify a connection.  ie. if two packets have the same tuple, they
  are in the same connection; if not, they are not. 

  We divide the structure along "manipulatable" and
  "non-manipulatable" lines, for the benefit of the NAT code.
*/

/* The protocol-specific manipulable parts of the tuple. */
union ip_conntrack_manip_proto
{
	/* Add other protocols here. */
	u_int16_t all;

	struct {
		u_int16_t port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
};

/* The manipulable part of the tuple. */
struct ip_conntrack_manip
{
	u_int32_t ip;
	union ip_conntrack_manip_proto u;
	u_int16_t pad; /* Must be set to 0 for memcmp. */
};

/* This contains the information to distinguish a connection. */
struct ip_conntrack_tuple
{
	struct ip_conntrack_manip src;

	/* These are the parts of the tuple which are fixed. */
	struct {
		u_int32_t ip;
		union {
			/* Add other protocols here. */
			u_int16_t all;

			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
		} u;

		/* The protocol. */
		u_int16_t protonum;
	} dst;
};

#ifdef __KERNEL__

#include <linux/types.h>
#include "struct_reserve.h"
#include "common/lockhelp.h"

/* You need this to examine/alter ip_conntrack hash, and manip refcnts.
 * Also protects all the other internal lists. */
DECLARE_LOCK_EXTERN(ip_conntrack_lck);
#define ASSERT_LOCK() MUST_BE_LOCKED(&ip_conntrack_lck)
#include "common/listhelp.h"

#ifdef DEBUG_CONNTRACK
#define IP_CONNTRACK_ASSERT(x)						 \
do {									 \
	if (!(x))							 \
		/* Wooah!  I'm tripping my conntrack in a frenzy of	 \
		   netplay... */					 \
		printk("ASSERT: %s:%i(%s)\n",				 \
		       __FILE__, __LINE__, __FUNCTION__);		 \
} while(0);
#define DEBUGP printk
#else
#define IP_CONNTRACK_ASSERT(x)
#define DEBUGP(format, args...)
#endif

/* Play with dynamic later; meanwhile if it's good enough for DaveM's
   TCP code... */
#define IP_CONNTRACK_HTABLE_SIZE 64

extern struct list_head ip_conntrack_hash[IP_CONNTRACK_HTABLE_SIZE];

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) (&(h)->ctrack->tuplehash[1] == (h))

enum ip_conntrack_info
{
	/* Not part of an established, and can't be used to make a new
	   one; eg. bogus ICMP error.  Conntrack pointer invalid. */
	IP_CT_INVALID = -1,
	/* Started a new connection to track (only in first
           direction); may be a retransmission. */
	IP_CT_NEW,

	/* Part of an established connection (either direction). */
	IP_CT_ESTABLISHED,

	/* Like NEW, but related to an existing connection, or ICMP error 
	   (in either direction). */
	IP_CT_RELATED,

	/* Specifies a related connection: in established conn
	   (either direction). */
	IP_CT_ESTABLISHED_SPECIFIES_RELATED,

	IP_CT_NUMBER
};

/* We set nfmark in the skb to point to one of the offsets[].  From this
   we can figure out the relation of the skb to the conntrack;
   effectively, we're encoding the relation in the pointer itself.
*/

/* Connections have two entries in the hash table: one for each way */
struct ip_conntrack_tuple_hash
{
	struct list_head list;

	struct ip_conntrack_tuple tuple;

	/* If it's dead, ignore it in the hash. */
	int dead;

	/* this == &ctrack->tuplehash[DIRECTION(this)]. */
	struct ip_conntrack *ctrack;

	/* Table of offsets to the start of the tuplehash. */
	u_int8_t offsets[IP_CT_NUMBER];
};

/* Return conntrack_info and tuple hash given skb. */
enum ip_conntrack_info
ip_conntrack_findinfo(struct sk_buff *skb, struct ip_conntrack_tuple_hash **h);

enum ip_conntrack_status {
	/* We haven't seen a reply packet yet. */
	IPS_AWAITING_REPLY,

	/* We've seen packets both ways. */
	IPS_BOTH_WAYS,

	/* It's an expected connection, but we haven't seen a reply
           packet yet. */
	IPS_EXPECTED_AWAITING_REPLY
};

struct ip_conntrack_expect
{
	/* Internal linked list */
	struct list_head list;

	/* We expect this tuple, but DON'T CARE ABOUT THE SOURCE
	   per-protocol part. */
	struct ip_conntrack_tuple tuple;
};

struct ip_conntrack
{
	/* These are my tuples; original and reply */
	struct ip_conntrack_tuple_hash tuplehash[2];

	enum ip_conntrack_status status;

	/* Usage count. */
	unsigned int use;

	/* Timer function; deletes us when we go off. */
	struct timer_list timeout;

	/* If we're expecting another related connection, this will be
           in expected linked list. */
	struct ip_conntrack_expect expected;

	/* Helper, if any. */
	struct ip_conntrack_helper *helper;
};

extern void ip_conntrack_get(struct ip_conntrack *ct);
extern void ip_conntrack_put(struct ip_conntrack *ct);

struct ip_conntrack_protocol
{
	/* Next pointer. */
	struct list_head list;

	/* Protocol number. */
	u_int8_t proto;

	/* Protocol name */
	const char *name;

	/* Try to fill in the third arg; return true if possible. */
	int (*pkt_to_tuple)(const void *datah, size_t datalen,
			    struct ip_conntrack_tuple *tuple);

	/* Invert the per-proto part of the tuple: ie. turn xmit into reply. 
	 * Some packets can't be inverted: return 0 in that case.
	 */
	int (*invert_tuple)(struct ip_conntrack_tuple *inverse,
			    const struct ip_conntrack_tuple *orig);

	/* Print out the per-protocol part of the tuple. */
	unsigned int (*print_tuple)(char *buffer,
				    const struct ip_conntrack_tuple *);

	/* Print out the private part of the conntrack. */
	unsigned int (*print_conntrack)(char *buffer,
					const struct ip_conntrack *);

	/* Returns verdict for packet, and may modify conntracktype */
	unsigned int (*established)(struct ip_conntrack *conntrack,
				    struct iphdr *iph, size_t len,
				    int direction, 
				    enum ip_conntrack_info *conntrackinfo);

	/* Called when a new connection for this protocol found. */
	void (*new)(struct ip_conntrack *conntrack,
		    struct iphdr *iph, size_t len,
		    enum ip_conntrack_info *conntrackinfo);

	/* Module (if any) which this is connected to. */
	struct module *me;

	/* Time (in HZ) until expiry.  0 = no expiry. */
	unsigned long expiry;
};

/* Alter reply tuple. */
extern void 
ip_conntrack_alter_reply(struct ip_conntrack *conntrack, 
			 const struct ip_conntrack_tuple *newreply);

/* Find a connection corresponding to a tuple (ignoring any belonging
   to the given conntrack) */
extern struct ip_conntrack_tuple_hash *
ip_conntrack_find(const struct ip_conntrack_tuple *tuple,
		  const struct ip_conntrack *ignored_conntrack);

/* Delete a conntrack. */
extern void ip_conntrack_delete(struct ip_conntrack *conntrack);

/* Expire in this direction.  If expired in both dirs, delete. */
extern void ip_conntrack_delete_maybe(struct ip_conntrack *conntrack, int direction);

/* Add a related connection. */
extern int ip_conntrack_expect_related(const struct ip_conntrack_tuple *tuple,
				       struct ip_conntrack *related_to);

extern struct module *ip_conntrack_module;

extern int invert_tuplepr(struct ip_conntrack_tuple *inverse,
			  const struct ip_conntrack_tuple *orig);

/* Call me when a conntrack is destroyed. */
extern void (*ip_conntrack_destroyed)(struct ip_conntrack *conntrack);

extern struct list_head expect_list;

EXTERN_DECLARE_STRUCT_RESERVABLE(ip_conntrack);
#endif /* __KERNEL__ */
#endif /* _IP_CONNTRACK_H */
