/* Kernel module to control the rate
 *
 * Jrme de Vivie   <devivie@info.enserb.u-bordeaux.fr>
 * Herv Eychenne   <eychenne@info.enserb.u-bordeaux.fr>
 *
 * 2 September 1999: Changed from the target RATE to the match
 *                   `limit', removed logging.  Did I mention that
 *                   Alexey is a fucking genius?
 *                   Rusty Russell (rusty@rustcorp.com.au).  */
#include <linux/module.h>
#include <linux/skbuff.h>
#include "packet-filter/kernel/ip_tables.h"
#include "packet-filter/extentions/ipt_limit.h"

#include <linux/spinlock.h>
EXPORT_NO_SYMBOLS;

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

/* The algorithm used is the Simple Token Bucket Filter (TBF)
 * see net/sched/sch_tbf.c in the linux source tree
 */

static spinlock_t limit_lock = SPIN_LOCK_UNLOCKED;

struct ipt_kern_rate {
	unsigned long prev;
	u_int32_t credit;

	/* These remain fixed. */
	u_int32_t credit_cap, cost;
};

/* Rusty: This is my (non-mathematically-inclined) understanding of
   this algorithm.  The `average rate' in jiffies becomes your initial
   amount of credit `credit' and the most credit you can ever have
   `credit_cap'.  The `peak rate' becomes the cost of passing the
   test, `cost'.

   `prev' tracks the last packet hit: you gain one credit per jiffy.
   If you get credit balance more than this, the extra credit is
   discarded.  Every time the match passes, you lose `cost' credits;
   if you don't have that many, the test fails.

   See Alexey's formal explanation in net/sched/sch_tbf.c. 

   To avoid underflow, we multiply by 128 (ie. you get 128 credits per
   jiffy).  Hence a cost of 2^32-1, means one pass per 32768 seconds
   at 1024HZ (or one every 9 hours).  A cost of 1 means 12800 passes
   per second at 100HZ.  */

static int 
ipt_limit_match(struct sk_buff *skb,
		const char *indev,
		const char *outdev,
		const union ipt_matchinfo *minfo,
		int offset,
		const void *hdr,
		u_int16_t datalen,
		int *hotdrop)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)minfo->other;
	struct ipt_kern_rate *kr = r->kr;
	unsigned long flags;
	unsigned long now = jiffies;
	
	spin_lock_irqsave(&limit_lock, flags);
	kr->credit += (now - xchg(&kr->prev, now)) * 128;
	if (kr->credit > kr->credit_cap)
		kr->credit = kr->credit_cap;

	if (kr->credit >= kr->cost) {
		/* We're not limited. */
		kr->credit -= kr->cost;
		spin_unlock_irqrestore(&limit_lock, flags);
		return 1;
	}

       	spin_unlock_irqrestore(&limit_lock, flags);
	return 0;
}

/* Precision saver. */
static u_int32_t
user2credits(u_int32_t user)
{
	/* If multiplying by HZ overflows... */
	if ((user * HZ) / HZ != user) {
		/* If multiplying by 128 overflows... */
		if ((user * 128) / 128 != user) {
			/* Divide first. */
			return (user / IPT_LIMIT_SCALE) * HZ * 128;
		}
		return (user * 128) / IPT_LIMIT_SCALE * HZ;
	}
	return (user * HZ * 128) / IPT_LIMIT_SCALE;
}

static int 
ipt_limit_checkentry(const struct ipt_ip *ip,
		     const union ipt_matchinfo *matchinfo)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)matchinfo->other;
	
	/* Check for overflow. */
	if (r->burst == 0 || r->avg * r->burst / r->burst != r->avg)
		return 0;

	r->kr = kmalloc(sizeof(struct ipt_kern_rate), 
#ifdef ALWAYS_IN_INTERRUPT
			GFP_ATOMIC
#else
			GFP_KERNEL
#endif
		);
	if (!r->kr)
		return 0;

	/* User avg in seconds * IPT_LIMIT_SCALE: convert to jiffies *
	   128. */
	*r->kr = ((struct ipt_kern_rate )
		  { 0,
		    user2credits(r->avg * r->burst),       /* Credits full. */
		    user2credits(r->avg * r->burst),       /* Credits cap. */
		    user2credits(r->avg) });               /* Cost. */
	return 1;
}

static void
ipt_limit_deleteentry(const union ipt_matchinfo *minfo)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)minfo->other;
	kfree(r->kr);
}

static struct ipt_match ipt_limit_reg
= { { NULL, NULL }, "limit", ipt_limit_match, ipt_limit_checkentry,
    ipt_limit_deleteentry, &__this_module };

int __init init(void)
{
	if (ipt_register_match(&ipt_limit_reg))
		return -EINVAL;
	return 0;
}

void __exit cleanup(void)
{
	ipt_unregister_match(&ipt_limit_reg);
}

module_init(init);
module_exit(cleanup);
