/* Shared library add-on to iptables to add TCP support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include "packet-filter/userspace/iptables.h"
#include "packet-filter/extentions/ipt_tcp.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"TCP v%s options:\n"
" --tcp-flags [!] mask comp	match when TCP flags & mask == comp\n"
"				(Flags: SYN ACK FIN RST URG PSH ALL NONE)\n"
"[!] --syn			match when only SYN flag set\n"
"				(equivalent to --tcp-flags SYN,RST,ACK SYN)\n"
" --source-port [!] port[:port]\n"
" --sport ...\n"
"				match source port(s)\n"
" --destination-port [!] port[:port]\n"
" --dport ...\n"
"				match destination port(s)\n"
" --tcp-option [!] number       match if TCP option set\n\n",
NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "source-port", 1, 0, '1' },
	{ "sport", 1, 0, '1' }, /* synonym */
	{ "destination-port", 1, 0, '2' },
	{ "dport", 1, 0, '2' }, /* synonym */
	{ "syn", 0, 0, '3' },
	{ "tcp-flags", 1, 0, '4' },
	{ "tcp-option", 1, 0, '5' },
	{0}
};

static int
service_to_port(const char *name)
{
	struct servent *service;

	if ((service = getservbyname(name, "tcp")) != NULL)
		return ntohs((unsigned short) service->s_port);

	return -1;
}

static u_int16_t
parse_tcp_port(const char *port)
{
	int portnum;

	if ((portnum = string_to_number(port, 0, 65535)) != -1 ||
	    (portnum = service_to_port(port)) != -1)
		return (u_int16_t)portnum;

	exit_error(PARAMETER_PROBLEM,
		   "invalid TCP port/service `%s' specified", port);
}

static void
parse_tcp_ports(const char *portstring, u_int16_t *ports)
{
	char *buffer;
	char *cp;

	buffer = strdup(portstring);
	if ((cp = strchr(buffer, ':')) == NULL)
		ports[0] = ports[1] = parse_tcp_port(buffer);
	else {
		*cp = '\0';
		cp++;

		ports[0] = buffer[0] ? parse_tcp_port(buffer) : 0;
		ports[1] = cp[0] ? parse_tcp_port(cp) : 0xFFFF;
	}
	free(buffer);
}

struct tcp_flag_names {
	const char *name;
	unsigned int flag;
};

static struct tcp_flag_names tcp_flag_names[]
= { { "FIN", 0x01 },
    { "SYN", 0x02 },
    { "RST", 0x04 },
    { "PSH", 0x08 },
    { "ACK", 0x10 },
    { "URG", 0x20 },
    { "ALL", 0x3F },
    { "NONE", 0 },
};

static unsigned int
parse_tcp_flag(const char *flags)
{
	unsigned int ret = 0;
	char *ptr;
	char *buffer;

	buffer = strdup(flags);

	for (ptr = strtok(buffer, ","); ptr; ptr = strtok(NULL, ",")) {
		unsigned int i;
		for (i = 0;
		     i < sizeof(tcp_flag_names)/sizeof(struct tcp_flag_names);
		     i++) {
			if (strcasecmp(tcp_flag_names[i].name, ptr) == 0) {
				ret |= tcp_flag_names[i].flag;
				break;
			}
		}
		if (i == sizeof(tcp_flag_names)/sizeof(struct tcp_flag_names))
			exit_error(PARAMETER_PROBLEM,
				   "Unknown TCP flag `%s'", buffer);
		}

	free(buffer);
	return ret;
}

static void
parse_tcp_flags(struct ipt_entry *entry,
		const char *mask,
		const char *cmp,
		int invert)
{
	entry->matchinfo.tcp.flg_mask = parse_tcp_flag(mask);
	entry->matchinfo.tcp.flg_cmp = parse_tcp_flag(cmp);

	if (invert)
		entry->matchinfo.tcp.invflags |= IPT_TCP_INV_FLAGS;
}

static void
parse_tcp_option(const char *option, u_int8_t *result)
{
	int ret;

	ret = string_to_number(option, 1, 255);
	if (ret == -1)
		exit_error(PARAMETER_PROBLEM, "Bad TCP option `%s'", option);

	*result = (u_int8_t)ret;
}

/* Initialize per-matchinfo part on first call. */
static void
init(union ipt_matchinfo *matchinfo, unsigned int *nfcache)
{
	matchinfo->tcp = ((struct ipt_tcp) { { 0, 0xFFFF },
					 { 0, 0xFFFF },
					 0,
					 0,
					 0,
					 0 });
}

#define TCP_SRC_PORTS 0x01
#define TCP_DST_PORTS 0x02
#define TCP_FLAGS 0x04
#define TCP_OPTION	0x08

/* Function which parses command options; returns true if it
   ate an option. */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      struct ipt_entry *entry)
{
	switch (c) {
	case '1':
		if (*flags & TCP_SRC_PORTS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--source-port' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		parse_tcp_ports(argv[optind-1], entry->matchinfo.tcp.spts);
		if (invert)
			entry->matchinfo.tcp.invflags |= IPT_TCP_INV_SRCPT;
		*flags |= TCP_SRC_PORTS;
		entry->nfcache |= NFC_IP_SRC_PT;
		break;

	case '2':
		if (*flags & TCP_DST_PORTS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--destination-port' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		parse_tcp_ports(argv[optind-1], entry->matchinfo.tcp.dpts);
		if (invert)
			entry->matchinfo.tcp.invflags |= IPT_TCP_INV_DSTPT;
		*flags |= TCP_DST_PORTS;
		entry->nfcache |= NFC_IP_DST_PT;
		break;

	case '3':
		if (*flags & TCP_FLAGS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one of `--syn' or `--tcp-flags' "
				   " allowed");
		parse_tcp_flags(entry, "SYN,RST,ACK", "SYN", invert);
		*flags |= TCP_FLAGS;
		entry->nfcache |= NFC_IP_TCPFLAGS;
		break;

	case '4':
		if (*flags & TCP_FLAGS)
			exit_error(PARAMETER_PROBLEM,
				   "Only one of `--syn' or `--tcp-flags' "
				   " allowed");
		if (check_inverse(optarg, &invert))
			optind++;

		if (!argv[optind]
		    || argv[optind][0] == '-' || argv[optind][0] == '!')
			exit_error(PARAMETER_PROBLEM,
				   "--tcp-flags requires two args.");

		parse_tcp_flags(entry, optarg, argv[optind++], invert);
		*flags |= TCP_FLAGS;
		entry->nfcache |= NFC_IP_TCPFLAGS;
		break;

	case '5':
		if (*flags & TCP_OPTION)
			exit_error(PARAMETER_PROBLEM,
				   "Only one `--tcp-option' allowed");
		if (check_inverse(optarg, &invert))
			optind++;
		parse_tcp_option(argv[optind-1], &entry->matchinfo.tcp.option);
		if (invert)
			entry->matchinfo.tcp.invflags |= IPT_TCP_INV_OPTION;
		*flags |= TCP_OPTION;
		entry->nfcache |= NFC_IP_PROTO_UNKNOWN;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; we don't care. */
static void
final_check(unsigned int flags)
{
}

static char *
port_to_service(int port)
{
	struct servent *service;

	if ((service = getservbyport(htons(port), "tcp")))
		return service->s_name;

	return NULL;
}

static void
print_port(u_int16_t port, int numeric)
{
	char *service;

	if (numeric || (service = port_to_service(htons(port))) == NULL)
		printf("%u", port);
	else
		printf("%s", service);
}

static void
print_ports(const char *name, u_int16_t min, u_int16_t max,
	    int invert, int numeric)
{
	const char *inv = invert ? "!" : "";

	if (min != 0 || max != 0xFFFF || invert) {
		printf("%s", name);
		if (min == max) {
			printf(":%s", inv);
			print_port(min, numeric);
		} else {
			printf("s:%s", inv);
			print_port(min, numeric);
			printf(":");
			print_port(max, numeric);
		}
		printf(" ");
	}
}

static void
print_option(u_int8_t option, int invert, int numeric)
{
	if (option || invert)
		printf("option=%s%u ", invert ? "!" : "", option);
}

static void
print_tcpf(u_int8_t flags)
{
	int sole_flag = 1;

	do {
		unsigned int i;

		/* Terminates because last flag is 0 */
		for (i = 0; !(flags & tcp_flag_names[i].flag); i++);

		if (!sole_flag)
			printf(",");
		printf("%s", tcp_flag_names[i].name);
		sole_flag = 0;

		flags &= ~tcp_flag_names[i].flag;
	} while (flags);
}

static void
print_flags(u_int8_t mask, u_int8_t cmp, int invert, int numeric)
{
	if (mask || invert) {
		printf("Flags:%s", invert ? "!" : "");
		if (numeric)
			printf("0x02%X/0x02%X ", mask, cmp);
		else {
			print_tcpf(cmp);
			printf("/");
			print_tcpf(mask);
		}
	}
}

/* Prints out the union ipt_matchinfo. */
static void
print(const union ipt_matchinfo *info, int numeric)
{
	printf("tcp ");
	print_ports("spt", info->tcp.spts[0], info->tcp.spts[1],
		    info->tcp.invflags & IPT_TCP_INV_SRCPT,
		    numeric);
	print_ports("dpt", info->tcp.dpts[0], info->tcp.dpts[1],
		    info->tcp.invflags & IPT_TCP_INV_DSTPT,
		    numeric);
	print_option(info->tcp.option,
		     info->tcp.invflags & IPT_TCP_INV_OPTION,
		     numeric);
	print_flags(info->tcp.flg_mask, info->tcp.flg_cmp,
		    info->tcp.invflags & IPT_TCP_INV_FLAGS,
		    numeric);
	if (info->tcp.invflags & ~IPT_TCP_INV_MASK)
		printf("Unknown invflags: 0x%X ",
		       info->tcp.invflags & ~IPT_TCP_INV_MASK);
}

struct iptables_match tcp
= { NULL,
    "tcp",
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    opts };

void
_init(void)
{
	register_match(&tcp);
}
